/* -*-c-*-
 *
 * Dump custom Lisp images for faster script execution
 *
 * (c) 2020 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Runlisp, a tool for invoking Common Lisp scripts.
 *
 * Runlisp is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * Runlisp is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Runlisp.  If not, see <https://www.gnu.org/licenses/>.
 */

/*----- Header files ------------------------------------------------------*/

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/wait.h>

#include "common.h"
#include "lib.h"
#include "mdwopt.h"

/*----- Static data -------------------------------------------------------*/

#define MAXLINE 16384u
struct linebuf {
  int fd;
  char *buf;
  unsigned off, len;
};

enum {
  JST_READY,
  JST_RUN,
  JST_DEAD,
  JST_NSTATE
};

struct job {
  struct treap_node _node;
  struct job *next;
  struct argv av;
  unsigned st;
  FILE *log;
  pid_t kid;
  int exit;
  struct linebuf out, err;
};
#define JOB_NAME(job) TREAP_NODE_KEY(job)
#define JOB_NAMELEN(job) TREAP_NODE_KEYLEN(job)

static struct treap jobs = TREAP_INIT;
static struct job *job_ready, *job_run, *job_dead;
static unsigned nrun, maxrun = 1;
static int rc = 0;
static int nullfd;

static int sig_pipe[2] = { -1, -1 };
static sigset_t caught, pending;
static int sigloss = -1;

static unsigned flags = 0;
#define AF_BOGUS 0x0001u
#define AF_SETCONF 0x0002u
#define AF_DRYRUN 0x0004u
#define AF_ALL 0x0008u
#define AF_FORCE 0x0010u
#define AF_CHECKINST 0x0020u

/*----- Main code ---------------------------------------------------------*/

static PRINTF_LIKE(1, 2) void bad(const char *msg, ...)
  { va_list ap; va_start(ap, msg); vmoan(msg, ap); va_end(ap); rc = 2; }

static const char *tmpdir;

static void set_tmpdir(void)
{
  struct dstr d = DSTR_INIT;
  size_t n;
  unsigned i;

  dstr_putf(&d, "%s/runlisp.%d.", my_getenv("TMPDIR", "/tmp"), getpid());
  i = 0; n = d.len;
  for (;;) {
    d.len = n; dstr_putf(&d, "%d", rand());
    if (!mkdir(d.p, 0700)) break;
    else if (errno != EEXIST)
      lose("failed to create temporary directory `%s': %s",
	   d.p, strerror(errno));
    else if (++i >= 32) {
      dstr_puts(&d, "???");
      lose("failed to create temporary directory `%s': too many attempts",
	   d.p);
    }
  }
  tmpdir = xstrndup(d.p, d.len); dstr_release(&d);
}

static void recursive_delete_(struct dstr *dd)
{
  size_t n = dd->len;
  DIR *dir;
  struct dirent *d;

  dd->p[n] = 0;
  dir = opendir(dd->p);
  if (!dir)
    lose("failed to open directory `%s' for cleanup: %s",
	 dd->p, strerror(errno));

  dd->p[n++] = '/';
  for (;;) {
    d = readdir(dir); if (!d) break;
    if (d->d_name[0] == '.' && (!d->d_name[1] ||
				(d->d_name[1] == '.' && !d->d_name[2])))
      continue;
    dd->len = n; dstr_puts(dd, d->d_name);
    if (!unlink(dd->p));
    else if (errno == EISDIR) recursive_delete_(dd);
    else lose("failed to delete file `%s': %s", dd->p, strerror(errno));
  }
  closedir(dir);
  dd->p[--n] = 0;
  if (rmdir(dd->p))
    lose("failed to delete directory `%s': %s", dd->p, strerror(errno));
}

static void recursive_delete(const char *path)
{
  struct dstr d = DSTR_INIT;
  dstr_puts(&d, path); recursive_delete_(&d); dstr_release(&d);
}

static void cleanup(void)
  { if (tmpdir) { recursive_delete(tmpdir); tmpdir = 0; } }

static int configure_fd(const char *what, int fd, int nonblock, int cloexec)
{
  int fl, nfl;

  if (nonblock != -1) {
    fl = fcntl(fd, F_GETFL); if (fl < 0) goto fail;
    if (nonblock) nfl = fl | O_NONBLOCK;
    else nfl = fl&~O_NONBLOCK;
    if (fl != nfl && fcntl(fd, F_SETFL, nfl)) goto fail;
  }

  if (cloexec != -1) {
    fl = fcntl(fd, F_GETFD); if (fl < 0) goto fail;
    if (cloexec) nfl = fl | FD_CLOEXEC;
    else nfl = fl&~FD_CLOEXEC;
    if (fl != nfl && fcntl(fd, F_SETFD, nfl)) goto fail;
  }

  return (0);

fail:
  bad("failed to configure %s descriptor: %s", what, strerror(errno));
  return (-1);
}

static void handle_signal(int sig)
{
  sigset_t old;
  char x = '!';

  sigprocmask(SIG_BLOCK, &caught, &old);
  sigaddset(&pending, sig);
  sigprocmask(SIG_SETMASK, &old, 0);

  DISCARD(write(sig_pipe[1], &x, 1));
}

#define JF_QUIET 1u
static void add_job(struct job ***tail_inout, unsigned f,
		    const char *name, size_t len)
{
  struct job *job;
  struct treap_path path;
  struct config_section *sect;
  struct config_var *dump_var, *cmd_var;
  struct dstr d = DSTR_INIT;
  struct argv av = ARGV_INIT;
  unsigned fef;

  job = treap_probe(&jobs, name, len, &path);
  if (job) {
    if (verbose >= 2) {
      moan("ignoring duplicate Lisp `%s'", JOB_NAME(job));
      return;
    }
  }

  sect = config_find_section_n(&config, 0, name, len);
  if (!sect) lose("unknown Lisp implementation `%.*s'", (int)len, name);
  name = CONFIG_SECTION_NAME(sect);
  dump_var = config_find_var(&config, sect, 0, "dump-image");
  if (!dump_var) {
    if (!(f&JF_QUIET))
      lose("don't know how to dump images for Lisp implementation `%s'",
	   name);
    goto end;
  }
  cmd_var = config_find_var(&config, sect, 0, "command");
  if (!cmd_var)
    lose("no `command' defined for Lisp implementation `%s'", name);

  config_subst_split_var(&config, sect, dump_var, &av);
  if (!av.n) lose("empty command for Lisp implementation `%s'", name);

  if (flags&AF_CHECKINST) {
    dstr_reset(&d);
    fef = (verbose >= 2 ? FEF_VERBOSE : 0);
    config_subst_var(&config, sect, cmd_var, &d);
    if (!found_in_path_p(d.p, fef) ||
	(STRCMP(d.p, !=, av.v[0]) && !found_in_path_p(av.v[0], fef))) {
      if (verbose >= 2) moan("skipping Lisp implementation `%s'", name);
      goto end;
    }
  }

  if (!(flags&AF_FORCE)) {
    dstr_reset(&d);
    config_subst_string(&config, sect, "<internal>", "${@IMAGE}", &d);
    if (!access(d.p, F_OK)) {
      if (verbose >= 2)
	moan("image `%s' already exists: skipping `%s'", d.p, name);
      goto end;
    }
  }

  job = xmalloc(sizeof(*job));
  job->st = JST_READY;
  job->kid = -1;
  job->out.fd = -1; job->out.buf = 0;
  job->err.fd = -1; job->err.buf = 0;
  job->av = av; argv_init(&av);
  treap_insert(&jobs, &path, &job->_node, name, len);
  **tail_inout = job; *tail_inout = &job->next;
end:
  dstr_release(&d); argv_release(&av);
}

static void release_job(struct job *job)
{
  if (job->kid > 0) kill(job->kid, SIGKILL); /* ?? */
  if (job->log && job->log != stdout) fclose(job->log);
  free(job->out.buf); if (job->out.fd >= 0) close(job->out.fd);
  free(job->err.buf); if (job->err.fd >= 0) close(job->err.fd);
  free(job);
}

static void finish_job(struct job *job)
{
  char buf[16483];
  size_t n;
  int ok = 0;

  fprintf(job->log, "%-13s > ", JOB_NAME(job));
  if (WIFEXITED(job->exit)) {
    if (!WEXITSTATUS(job->exit))
      { fputs("completed successfully\n", job->log); ok = 1; }
    else
      fprintf(job->log, "failed with exit status %d\n",
	      WEXITSTATUS(job->exit));
  } else if (WIFSIGNALED(job->exit))
    fprintf(job->log, "killed by signal %d (%s%s)", WTERMSIG(job->exit),
#if defined(HAVE_STRSIGNAL)
	strsignal(WTERMSIG(job->exit)),
#elif defined(HAVE_DECL_SYS_SIGLIST)
	sys_siglist[WTERMSIG(job->exit)],
#else
	"unknown signal",
#endif
#ifdef WCOREDUMP
	WCOREDUMP(job->exit) ? "; core dumped" :
#endif
	"");
  else
    fprintf(job->log, "exited with incomprehensible status %06o\n",
	    job->exit);

  if (!ok && verbose < 2) {
    rewind(job->log);
    for (;;) {
      n = fread(buf, 1, sizeof(buf), job->log);
      if (n) fwrite(buf, 1, n, stdout);
      if (n < sizeof(buf)) break;
    }
  }

  release_job(job);
}

static int find_newline(struct linebuf *buf, size_t *linesz_out)
{
  char *nl;

  if (buf->off + buf->len <= MAXLINE) {
    nl = memchr(buf->buf + buf->off, '\n', buf->len);
    if (nl) { *linesz_out = (nl - buf->buf) - buf->off; return (0); }
  } else {
    nl = memchr(buf->buf + buf->off, '\n', MAXLINE - buf->off);
    if (nl) { *linesz_out = (nl - buf->buf) - buf->off; return (0); }
    nl = memchr(buf->buf, '\n', buf->len - (MAXLINE - buf->off));
    if (nl)
      { *linesz_out = (nl - buf->buf) + (MAXLINE - buf->off); return (0); }
  }
  return (-1);
}

static void write_line(struct job *job, struct linebuf *buf,
		       size_t n, char marker, const char *tail)
{
  fprintf(job->log, "%-13s %c ", JOB_NAME(job), marker);
  if (buf->off + n <= MAXLINE)
    fwrite(buf->buf + buf->off, 1, n, job->log);
  else {
    fwrite(buf->buf + buf->off, 1, MAXLINE - buf->off, job->log);
    fwrite(buf->buf, 1, n - (MAXLINE - buf->off), job->log);
  }
  fputs(tail, job->log);
}

static void prefix_lines(struct job *job, struct linebuf *buf, char marker)
{
  struct iovec iov[2]; int niov;
  ssize_t n;
  size_t linesz;

  assert(buf->len < MAXLINE);
  if (!buf->off) {
    iov[0].iov_base = buf->buf + buf->len;
    iov[0].iov_len = MAXLINE - buf->len;
    niov = 1;
  } else if (buf->off + buf->len >= MAXLINE) {
    iov[0].iov_base = buf->buf + buf->off + buf->len - MAXLINE;
    iov[0].iov_len = MAXLINE - buf->len;
    niov = 1;
  } else {
    iov[0].iov_base = buf->buf + buf->off + buf->len;
    iov[0].iov_len = MAXLINE - (buf->off + buf->len);
    iov[1].iov_base = buf->buf;
    iov[1].iov_len = buf->off;
    niov = 1;
  }

  n = readv(buf->fd, iov, niov);
  if (n < 0) {
    if (errno == EAGAIN || errno == EWOULDBLOCK) return;
    lose("failed to read job `%s' output stream: %s",
	 JOB_NAME(job), strerror(errno));
  }
  buf->len += n;

  while (!find_newline(buf, &linesz)) {
    write_line(job, buf, linesz, marker, "\n");
    buf->len -= linesz + 1;
    buf->off += linesz + 1; if (buf->off >= MAXLINE) buf->off -= MAXLINE;
  }
  if (!buf->len)
    buf->off = 0;
  else if (buf->len == MAXLINE) {
    write_line(job, buf, MAXLINE, marker, " [...]\n");
    buf->off = buf->len = 0;
  }

  if (!n) {
    close(buf->fd); buf->fd = -1;
    if (buf->len)
      write_line(job, buf, buf->len, marker, " [missing final newline]\n");
  }
}

static void reap_children(void)
{
  struct job *job, **link;
  pid_t kid;
  int st;

  for (;;) {
    kid = waitpid(0, &st, WNOHANG);
    if (kid <= 0) break;
    for (link = &job_run; (job = *link); link = &job->next)
      if (job->kid == kid) goto found;
    moan("unexpected child process %d exited with status %06o", kid, st);
    continue;
  found:
    job->exit = st; job->st = JST_DEAD; job->kid = -1; nrun--;
    *link = job->next; job->next = job_dead; job_dead = job;
  }
  if (kid < 0 && errno != ECHILD)
    lose("failed to collect child process exit status: %s", strerror(errno));
}

static void check_signals(void)
{
  sigset_t old, pend;
  char buf[32];
  ssize_t n;

  sigprocmask(SIG_BLOCK, &caught, &old);
  pend = pending; sigemptyset(&pending);
  for (;;) {
    n = read(sig_pipe[0], buf, sizeof(buf));
    if (!n) lose("(internal) signal pipe closed!");
    if (n < 0) break;
  }
  if (errno != EAGAIN && errno != EWOULDBLOCK)
    lose("failed to read signal pipe: %s", strerror(errno));
  sigprocmask(SIG_SETMASK, &old, 0);

  if (sigismember(&pend, SIGINT)) sigloss = SIGINT;
  else if (sigismember(&pend, SIGHUP)) sigloss = SIGHUP;
  else if (sigismember(&pend, SIGTERM)) sigloss = SIGTERM;
  if (sigismember(&pend, SIGCHLD)) reap_children();
}

#define SIGF_IGNOK 1u
static void set_signal_handler(const char *what, int sig, unsigned f)
{
  struct sigaction sa, sa_old;

  sigaddset(&caught, sig);

  if (f&SIGF_IGNOK) {
    if (sigaction(sig, 0, &sa_old)) goto fail;
    if (sa_old.sa_handler == SIG_IGN) return;
  }

  sa.sa_handler = handle_signal;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_NOCLDSTOP;
  if (sigaction(sig, &sa, 0)) goto fail;

  return;

fail:
  lose("failed to set %s signal handler: %s", what, strerror(errno));
}

static NORETURN void job_child(struct job *job)
{
  try_exec(&job->av,
	   !(flags&AF_CHECKINST) && verbose >= 2 ? TEF_VERBOSE : 0);
  moan("failed to run `%s': %s", job->av.v[0], strerror(errno));
  _exit(2);
}

static void start_jobs(void)
{
  struct dstr d = DSTR_INIT;
  int p_out[2], p_err[2];
  struct job *job;
  pid_t kid;

  while (job_ready && nrun < maxrun) {
    job = job_ready; job_ready = job->next;
    p_out[0] = p_out[1] = p_err[0] = p_err[1] = -1;
    dstr_reset(&d); dstr_putf(&d, "%s/%s", tmpdir, JOB_NAME(job));
    if (mkdir(d.p, 0700)) {
      bad("failed to create working directory for job `%s': %s",
	  JOB_NAME(job), strerror(errno));
      goto fail;
    }
    if (verbose >= 2)
      job->log = stdout;
    else {
      dstr_puts(&d, "/log"); job->log = fopen(d.p, "w+");
      if (!job->log)
	lose("failed to open log file `%s': %s", d.p, strerror(errno));
    }
    if (pipe(p_out) || pipe(p_err)) {
      bad("failed to create pipes for job `%s': %s",
	  JOB_NAME(job), strerror(errno));
      goto fail;
    }
    if (configure_fd("job stdout pipe", p_out[0], 1, 1) ||
	configure_fd("job stdout pipe", p_out[1], 0, 1) ||
	configure_fd("job stderr pipe", p_err[0], 1, 1) ||
	configure_fd("job stderr pipe", p_err[1], 0, 1) ||
	configure_fd("log file", fileno(job->log), 1, 1))
      goto fail;
    job->out.buf = xmalloc(MAXLINE); job->out.off = job->out.len = 0;
    job->out.fd = p_out[0]; p_out[0] = -1;
    job->err.buf = xmalloc(MAXLINE); job->err.off = job->err.len = 0;
    job->err.fd = p_err[0]; p_err[0] = -1;
    dstr_reset(&d); argv_string(&d, &job->av);
    fprintf(job->log, "%-13s > starting %s\n", JOB_NAME(job), d.p);
    fflush(stdout);
    kid = fork();
    if (kid < 0) {
      bad("failed to fork process for job `%s': %s",
	  JOB_NAME(job), strerror(errno));
      goto fail;
    }
    if (!kid) {
      if (dup2(nullfd, 0) < 0 ||
	  dup2(p_out[1], 1) < 0 ||
	  dup2(p_err[1], 2) < 0)
	lose("failed to juggle job `%s' file descriptors: %s",
	     JOB_NAME(job), strerror(errno));
      job_child(job);
    }
    close(p_out[1]); close(p_err[1]);
    job->kid = kid;
    job->st = JST_RUN; job->next = job_run; job_run = job; nrun++;
    continue;
  fail:
    if (p_out[0] >= 0) close(p_out[0]);
    if (p_out[1] >= 0) close(p_out[1]);
    if (p_err[0] >= 0) close(p_err[0]);
    if (p_err[1] >= 0) close(p_err[1]);
    release_job(job);
  }
  dstr_release(&d);
}

static void version(FILE *fp)
  { fprintf(fp, "%s, runlisp version %s\n", progname, PACKAGE_VERSION); }

static void usage(FILE *fp)
{
  fprintf(fp, "\
usage: %s [-afnqv] [-c CONF] [-o [SECT:]VAR=VAL]\n\
	[-O FILE|DIR] [-j NJOBS] [LISP ...]\n",
	  progname);
}

static void help(FILE *fp)
{
  version(fp); fputc('\n', fp); usage(fp);
  fputs("\n\
Help options:\n\
  -h, --help			Show this help text and exit successfully.\n\
  -V, --version			Show version number and exit successfully.\n\
\n\
Diagnostics:\n\
  -n, --dry-run			Don't run run anything (useful with `-v').\n\
  -q, --quiet			Don't print warning messages.\n\
  -v, --verbose			Print informational messages (repeatable).\n\
\n\
Configuration:\n\
  -c, --config-file=CONF	Read configuration from CONF (repeatable).\n\
  -o, --set-option=[SECT:]VAR=VAL Set configuration variable (repeatable).\n\
\n\
Image dumping:\n\
  -O, --output=FILE|DIR		Store image(s) in FILE or DIR.\n\
  -a, --all-configured		Dump all implementations configured.\n\
  -f, --force			Dump images even if they already exist.\n\
  -i, --check-installed		Check Lisp systems exist before invoking.\n\
  -j, --jobs=NJOBS		Run up to NJOBS jobs in parallel.\n",
	fp);
}

int main(int argc, char *argv[])
{
  struct config_section_iter si;
  struct config_section *sect;
  struct config_var *var;
  const char *out = 0, *p, *q, *l;
  struct job *job, **tail, **link, *next;
  struct stat st;
  struct dstr d = DSTR_INIT;
  int i, fd, nfd, first;
  fd_set fd_in;

  static const struct option opts[] = {
    { "help",			0,		0,	'h' },
    { "version",		0,		0,	'V' },
    { "output",			OPTF_ARGREQ,	0,	'O' },
    { "all-configured",		0,		0,	'a' },
    { "config-file",		OPTF_ARGREQ,	0,	'c' },
    { "force",			OPTF_NEGATE,	0,	'f' },
    { "check-installed",	OPTF_NEGATE,	0,	'i' },
    { "jobs",			OPTF_ARGREQ,	0,	'j' },
    { "dry-run",		OPTF_NEGATE,	0,	'n' },
    { "set-option",		OPTF_ARGREQ,	0,	'o' },
    { "quiet",			0,		0,	'q' },
    { "verbose",		0,		0,	'v' },
    { 0,			0,		0,	0 }
  };

  set_progname(argv[0]);
  init_config();

  optprog = (/*unconst*/ char *)progname;
  for (;;) {
    i = mdwopt(argc - 1, argv + 1, "hVO:ac:f+i+j:n+o:qv", opts, 0, 0,
	       OPTF_NEGATION | OPTF_NOPROGNAME);
    if (i < 0) break;
    switch (i) {
      case 'h': help(stdout); exit(0);
      case 'V': version(stdout); exit(0);
      case 'O': out = optarg; break;
      case 'a': flags |= AF_ALL; break;
      case 'c': read_config_path(optarg, 0); flags |= AF_SETCONF; break;
      case 'f': flags |= AF_FORCE; break;
      case 'f' | OPTF_NEGATED: flags &= ~AF_FORCE; break;
      case 'i': flags |= AF_CHECKINST; break;
      case 'i' | OPTF_NEGATED: flags &= ~AF_CHECKINST; break;
      case 'j': maxrun = parse_int("number of jobs", optarg, 1, 65535); break;
      case 'n': flags |= AF_DRYRUN; break;
      case 'n' | OPTF_NEGATED: flags &= ~AF_DRYRUN; break;
      case 'o': if (set_config_var(optarg)) flags |= AF_BOGUS; break;
      case 'q': if (verbose) verbose--; break;
      case 'v': verbose++; break;
      default: flags |= AF_BOGUS; break;
    }
  }

  optind++;
  if ((flags&AF_ALL) ? optind < argc : optind >= argc) flags |= AF_BOGUS;
  if (flags&AF_BOGUS) { usage(stderr); exit(2); }

  if (!(flags&AF_SETCONF)) load_default_config();

  if (!out)
    config_set_var(&config, builtin, 0,
		   "@IMAGE", "${@CONFIG:image-dir}/${image-file}");
  else if (stat(out, &st) || !S_ISDIR(st.st_mode))
    config_set_var(&config, builtin, CF_LITERAL, "@IMAGE", out);
  else {
    config_set_var(&config, builtin, CF_LITERAL, "@%OUTDIR", out);
    config_set_var(&config, builtin, 0,
		   "@IMAGE", "${@BUILTIN:@%OUTDIR}/${image-file}");
  }

  atexit(cleanup);
  if (pipe(sig_pipe))
    lose("failed to create signal pipe: %s", strerror(errno));
  configure_fd("signal pipe (read end)", sig_pipe[0], 1, 1);
  configure_fd("signal pipe (write end)", sig_pipe[1], 1, 1);
  sigemptyset(&caught); sigemptyset(&pending);
  set_signal_handler("SIGTERM", SIGTERM, SIGF_IGNOK);
  set_signal_handler("SIGINT", SIGINT, SIGF_IGNOK);
  set_signal_handler("SIGHUP", SIGHUP, SIGF_IGNOK);
  set_signal_handler("SIGCHLD", SIGCHLD, 0);

  set_tmpdir();
  config_set_var(&config, builtin, CF_LITERAL, "@%TMPDIR", tmpdir);
  config_set_var(&config, builtin, 0,
		 "@TMPDIR", "${@BUILTIN:@%TMPDIR}/${@NAME}");

  if (verbose >= 5) dump_config();

  tail = &job_ready;
  if (!(flags&AF_ALL))
    for (i = optind; i < argc; i++)
      add_job(&tail, 0, argv[i], strlen(argv[i]));
  else {
    var = config_find_var(&config, toplevel, 0, "dump");
    if (!var)
      for (config_start_section_iter(&config, &si);
	   (sect = config_next_section(&si)); )
	add_job(&tail, JF_QUIET,
		CONFIG_SECTION_NAME(sect),
		CONFIG_SECTION_NAMELEN(sect));
    else {
      p = var->val; l = p + var->n;
      for (;;) {
	while (p < l && ISSPACE(*p)) p++;
	if (p >= l) break;
	q = p;
	while (p < l && !ISSPACE(*p) && *p != ',') p++;
	add_job(&tail, 0, q, p - q);
	if (p < l) p++;
      }
    }
  }
  *tail = 0;

  if (verbose >= 3) {
    dstr_reset(&d);
    first = 1;
    for (job = job_ready; job; job = job->next) {
      if (first) first = 0;
      else dstr_puts(&d, ", ");
      dstr_putf(&d, "`%s'", JOB_NAME(job));
    }
    if (first) dstr_puts(&d, "(none)");
    dstr_putz(&d);
    moan("dumping Lisps: %s", d.p);
  }

  if (flags&AF_DRYRUN) {
    for (job = job_ready; job; job = job->next) {
      if (try_exec(&job->av,
		   TEF_DRYRUN |
		     (verbose >= 2 && !(flags&AF_CHECKINST) ?
		       TEF_VERBOSE : 0)))
	rc = 2;
      else if (verbose >= 2)
	printf("%-13s > (not dumping `%s': dry run)\n",
		JOB_NAME(job), JOB_NAME(job));
    }
    return (rc);
  }

  for (;;) {
    fd = open("/dev/null", O_RDWR);
    if (fd < 0) lose("failed to open `/dev/null': %s", strerror(errno));
    if (fd > 2) { nullfd = fd; break; }
  }
  configure_fd("null fd", nullfd, 0, 1);

  for (;;) {
    start_jobs();
    if (!job_run && !job_dead) break;

#define SET_FD(dir, fd) do {						\
  int _fd = (fd);							\
									\
  FD_SET(_fd, &fd_##dir);						\
  if (_fd >= nfd) nfd = _fd + 1;					\
} while (0)

    FD_ZERO(&fd_in); nfd = 0;
    SET_FD(in, sig_pipe[0]);
    for (job = job_run; job; job = job->next) {
      if (job->out.fd >= 0) SET_FD(in, job->out.fd);
      if (job->err.fd >= 0) SET_FD(in, job->err.fd);
    }
    for (job = job_dead; job; job = job->next) {
      if (job->out.fd >= 0) SET_FD(in, job->out.fd);
      if (job->err.fd >= 0) SET_FD(in, job->err.fd);
    }

#undef SET_FD

    if (select(nfd, &fd_in, 0, 0, 0) < 0) {
      if (errno == EINTR) continue;
      else lose("select failed: %s", strerror(errno));
    }

    if (FD_ISSET(sig_pipe[0], &fd_in)) {
      check_signals();
      if (sigloss >= 0) {
	for (job = job_ready; job; job = next)
	  { next = job->next; release_job(job); }
	for (job = job_run; job; job = next)
	  { next = job->next; release_job(job); }
	for (job = job_dead; job; job = next)
	  { next = job->next; release_job(job); }
	break;
      }
    }

    for (job = job_run; job; job = job->next) {
      if (job->out.fd >= 0 && FD_ISSET(job->out.fd, &fd_in))
	prefix_lines(job, &job->out, '|');
      if (job->err.fd >= 0 && FD_ISSET(job->err.fd, &fd_in))
	prefix_lines(job, &job->err, '*');
    }
    for (link = &job_dead, job = *link; job; job = next) {
      next = job->next;
      if (job->out.fd >= 0 && FD_ISSET(job->out.fd, &fd_in))
	prefix_lines(job, &job->out, '|');
      if (job->err.fd >= 0 && FD_ISSET(job->err.fd, &fd_in))
	prefix_lines(job, &job->err, '*');
      if (job->out.fd >= 0 || job->err.fd >= 0) link = &job->next;
      else {  *link = next; finish_job(job); }
    }
  }

  check_signals();
  if (sigloss) { cleanup(); signal(sigloss, SIG_DFL); raise(sigloss); }

  return (rc);
}

/*----- That's all, folks -------------------------------------------------*/
